#ifndef ITC_H
#define ITC_H

#include <iostream>
#include <string>
#include <queue>
#include <vector>
#include <algorithm>
#include <cassert>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>

namespace itc {

typedef boost::function<void (void)> CallType;

struct ProcessingPolicy
{
	virtual void queue(CallType call) = 0;
	virtual void execute() = 0;
	virtual ~ProcessingPolicy() {}
};

class WaitAndExecute : public ProcessingPolicy
{
public:
	WaitAndExecute():
		m_mutex(), m_cond(), m_req(false), m_callQ()
	{}	

	void queue(CallType call)
	{
		boost::unique_lock<boost::mutex> lock(m_mutex);
		m_callQ.push(call);
		m_req = true;
		m_cond.notify_one();
	}
	
	void execute()
	{
		boost::unique_lock<boost::mutex> lock(m_mutex);
		
		// wait
		while (!m_req)
		{
			m_cond.wait(lock);
		}
		
		// execute
		while (!m_callQ.empty())
		{
			m_callQ.front()();
			m_callQ.pop();
		}
		m_req = false;	
	}	

private:
	boost::mutex m_mutex;
	boost::condition_variable m_cond;
	bool m_req;
	std::queue<CallType>  m_callQ;	
};

class ProcessAndExecute : public ProcessingPolicy
{
public:
	ProcessAndExecute() : m_mutex(), m_callQ()	{}
	
	void queue(CallType call)
	{
		boost::unique_lock<boost::mutex> lock(m_mutex);
		m_callQ.push(call);
	}
	
	void execute()
	{
		CallType call;
		bool req = false;
		
		// process
		process();
		
		// execute
		{
			boost::unique_lock<boost::mutex> lock(m_mutex);
			if (!m_callQ.empty())
			{
				call = m_callQ.front();
				m_callQ.pop();
				req = true;
			}
		}
		
		if (req)
			call();		
	}	

private:
	virtual void process() = 0;

	boost::mutex m_mutex;
	std::queue<CallType>  m_callQ;
};

class ITCThreadBase
{
public:
	ITCThreadBase(boost::shared_ptr<ProcessingPolicy> policy, const std::string name) :
		m_thr(), m_policy(policy),
		m_name(name), m_run(true), m_code(0)
	{
		m_thr = boost::thread(&ITCThreadBase::run, this);
	}

	const std::string& getName()
	{
		return m_name;
	}

	void quit(int code = 0)
	{
		queue(boost::bind(&ITCThreadBase::doQuit, this, code));
	}

	int join()
	{
		m_thr.join();
		return m_code;
	}

	void queue(CallType call)
	{
		m_policy->queue(call);
	}

private:
	void run()
	{
		while (m_run)
		{
			m_policy->execute();
		}
	}

	void doQuit(int code)
	{
		m_run = false;
		m_code = code;
	}

	boost::thread m_thr;
	boost::shared_ptr<ProcessingPolicy> m_policy;
	const std::string m_name;
	bool m_run;
	int m_code;
};

template <typename DerivedThread, typename PolicyType = WaitAndExecute>
struct ITCThread : public ITCThreadBase
{
	ITCThread(const std::string& name = "itcthr") :
		ITCThreadBase(boost::shared_ptr<ProcessingPolicy>(new PolicyType()), name)
	{}

	void doIt(void (DerivedThread::*mf)(void))
	{
		queue(boost::bind(mf, (DerivedThread*)this));
	}

	template <typename ArgType>
	void doIt(void (DerivedThread::*mf)(ArgType), ArgType arg)
	{
		queue(boost::bind(mf, (DerivedThread*)this, arg));
	}

	template <typename ArgType1, typename ArgType2>
	void doIt(void (DerivedThread::*mf)(ArgType1, ArgType2),
			  ArgType1 arg1, ArgType2 arg2)
	{
		queue(boost::bind(mf, (DerivedThread*)this, arg1, arg2));
	}

	template <typename ArgType1, typename ArgType2, typename ArgType3>
	void doIt(void (DerivedThread::*mf)(ArgType1, ArgType2, ArgType3),
			  ArgType1 arg1, ArgType2 arg2, ArgType3 arg3)
	{
		queue(boost::bind(mf, (DerivedThread*)this, arg1, arg2, arg3));
	}
};

template <typename Callee, typename PolicyType = WaitAndExecute>
class ITCAdaptor : public ITCThreadBase
{
public:
	ITCAdaptor(boost::shared_ptr<Callee> callee, const std::string& name = "itcapt") :
		ITCThreadBase(boost::shared_ptr<ProcessingPolicy>(new PolicyType()), name), m_callee(callee)
	{}

	void doIt(void (Callee::*mf)(void))
	{
		assert(m_callee != 0);
		queue(boost::bind(mf, m_callee));
	}

	template <typename ArgType>
	void doIt(void (Callee::*mf)(ArgType), ArgType arg)
	{
		assert(m_callee != 0);
		queue(boost::bind(mf, m_callee, arg));
	}

	template <typename ArgType1, typename ArgType2>
	void doIt(void (Callee::*mf)(ArgType1, ArgType2), 
			  ArgType1 arg1, ArgType2 arg2)
	{
		assert(m_callee != 0);
		queue(boost::bind(mf, m_callee, arg1, arg2));
	}

	template <typename ArgType1, typename ArgType2, typename ArgType3>
	void doIt(void (Callee::*mf)(ArgType1, ArgType2, ArgType3),
			  ArgType1 arg1, ArgType2 arg2, ArgType3 arg3)
	{
		assert(m_callee != 0);
		queue(boost::bind(mf, m_callee, arg1, arg2, arg3));
	}

private:
	boost::shared_ptr<Callee> m_callee;
};

template <typename PolicyType = WaitAndExecute>
struct FunctionITCAdaptor : public ITCThreadBase
{
	FunctionITCAdaptor(const std::string& name = "fnitcapt") :
		ITCThreadBase(boost::shared_ptr<ProcessingPolicy>(new PolicyType()), name)
	{}

    void doIt(void (*func)(void))
    {
        queue(func);
    }

    template <typename ArgType>
    void doIt(void (*func)(ArgType), ArgType arg)
    {
        queue(boost::bind(func, arg));
    }

    template <typename ArgType1, typename ArgType2>
    void doIt(void (*func)(ArgType1, ArgType2), ArgType1 arg1, ArgType2 arg2)
    {
        queue(boost::bind(func, arg1, arg2));
    }

    template <typename ArgType1, typename ArgType2, typename ArgType3>
    void doIt(void (*func)(ArgType1, ArgType2, ArgType3),
              ArgType1 arg1, ArgType2 arg2, ArgType3 arg3)
    {
        queue(boost::bind(func, arg1, arg2, arg3));
    }

	template <typename ArgType>
	struct DoItSlot
	{
		typedef void (FunctionITCAdaptor<PolicyType>::*type)(void (*)(ArgType), ArgType);
	};
	
};

class ITCSignal0
{
public:
	typedef boost::function<void ()> SlotType;
	typedef const SlotType& ConnectionType;

	ITCSignal0() : m_vec() {}

	template <typename C>
	ConnectionType connect(C* inst, void (C::*call)())
	{
		return connect(boost::bind(call, inst));
	}

	template <typename Caller, typename Callee = Caller>
	struct DoIt
	{
		typedef void (Callee::*CallType)();
		typedef void (Caller::*Type)(CallType);
	};

	template <typename Caller, typename Callee>
	ConnectionType connect(Caller* inst, 
						   typename DoIt<Caller, Callee>::CallType call)
	{
		return connect(boost::bind(typename DoIt<Caller,Callee>::Type(&Caller::doIt), inst, call));
	}

	template <typename C>
	struct FuncDoIt
	{
		typedef void (*CallType)();
		typedef void (C::*Type)(CallType);
	};

	template <typename C>
	ConnectionType connect(C* inst, typename FuncDoIt<C>::CallType call)
	{
		return connect(boost::bind(typename FuncDoIt<C>::Type(&C::doIt), inst, call));
	}

	ConnectionType connect(ConnectionType call)
	{
		m_vec.push_back(call);
		return m_vec[m_vec.size()-1];
	}

	void disconnect(ConnectionType call)
	{
		std::vector<SlotType>::iterator it = m_vec.begin();
		for (;it != m_vec.end(); ++it)
		{
			if (&(*it) == &call)
				break;
		}
		if (it != m_vec.end())
		{
			m_vec.erase(it);
		}
	}

	void operator()()
	{
		std::for_each(m_vec.begin(), m_vec.end(),
					  boost::mem_fn(&SlotType::operator()));
	}

private:
	std::vector<SlotType> m_vec;
};

template <typename ArgType>
class ITCSignal1
{
public:
	typedef boost::function<void (ArgType)> SlotType;
	typedef const SlotType& ConnectionType;
	
	ITCSignal1() : m_vec() {}

	template <typename C>
	ConnectionType connect(C* inst, void (C::*call)(ArgType))
	{
		return connect(boost::bind(call, inst, _1));
	}

	template <typename Caller, typename Callee = Caller>
	struct DoIt
	{
		typedef void (Callee::*CallType)(ArgType);
		typedef void (Caller::*Type)(CallType, ArgType);
	};

	template <typename Caller, typename Callee>
	ConnectionType connect(Caller* inst, 
						   typename DoIt<Caller, Callee>::CallType call)
	{
		return connect(boost::bind(typename DoIt<Caller,Callee>::Type(&Caller::template doIt<ArgType>), 
								   inst, call, _1));
	}
	
	template <typename C>
	struct FuncDoIt
	{
		typedef void (*CallType)(ArgType);
		typedef void (C::*Type)(CallType, ArgType);
	};
	
	template <typename C>
	ConnectionType connect(C* inst, typename FuncDoIt<C>::CallType call)
	{
		return connect(boost::bind(typename FuncDoIt<C>::Type(&C::template doIt<ArgType>), 
								   inst, call, _1));
	}

	ConnectionType connect(ConnectionType call)
	{
		m_vec.push_back(call);
		return m_vec[m_vec.size()-1];
	}
	
	void disconnect(ConnectionType call)
	{
		typename std::vector<SlotType>::iterator it = m_vec.begin();
		for (;it != m_vec.end(); ++it)
		{
			if (&(*it) == &call)
				break;
		}
		if (it != m_vec.end())
		{
			m_vec.erase(it);
		}
	}
	
	void operator()(ArgType arg)
	{
		std::for_each(m_vec.begin(), m_vec.end(),
					  boost::bind(&SlotType::operator(), _1, arg));
	}
	
private:
	std::vector<SlotType> m_vec;
};

template <typename ArgType1, typename ArgType2>
class ITCSignal2
{
public:
	typedef boost::function<void (ArgType1, ArgType2)> SlotType;
	typedef const SlotType& ConnectionType;

	ITCSignal2() : m_vec() {}

	template <typename C>
	ConnectionType connect(C* inst, void (C::*call)(ArgType1, ArgType2))
	{
		return connect(boost::bind(call, inst, _1, _2));
	}

	template <typename Caller, typename Callee = Caller>
	struct DoIt
	{
		typedef void (Callee::*CallType)(ArgType1, ArgType2);
		typedef void (Caller::*Type)(CallType, ArgType1, ArgType2);
	};

	template <typename Caller, typename Callee>
	ConnectionType connect(Caller* inst, 
						   typename DoIt<Caller, Callee>::CallType call)
	{
		return connect(boost::bind(typename DoIt<Caller, Callee>::Type(
										&Caller::template doIt<ArgType1, ArgType2>), 
								   inst, call, _1, _2));
	}

	template <typename C>
	struct FuncDoIt
	{
		typedef void (*CallType)(ArgType1, ArgType2);
		typedef void (C::*Type)(CallType, ArgType1, ArgType2);
	};

	template <typename C>
	ConnectionType connect(C* inst, typename FuncDoIt<C>::CallType call)
	{
		return connect(boost::bind(typename FuncDoIt<C>::Type(
										&C::template doIt<ArgType1, ArgType2>), 
								   inst, call, _1, _2));
	}
	
	ConnectionType connect(ConnectionType call)
	{
		m_vec.push_back(call);
		return m_vec[m_vec.size()-1];
	}

	void disconnect(ConnectionType call)
	{
		typename std::vector<SlotType>::iterator it = m_vec.begin();
		for (;it != m_vec.end(); ++it)
		{
			if (&(*it) == &call)
				break;
		}
		if (it != m_vec.end())
		{
			m_vec.erase(it);
		}
	}

	void operator()(ArgType1 arg1, ArgType2 arg2)
	{
		std::for_each(m_vec.begin(), m_vec.end(),
					  boost::bind(&SlotType::operator(), _1, arg1, arg2));
	}

private:
	std::vector<SlotType> m_vec;
};

template <typename ArgType1, typename ArgType2, typename ArgType3>
class ITCSignal3
{
public:
	typedef boost::function<void (ArgType1, ArgType2, ArgType3)> SlotType;
	typedef const SlotType& ConnectionType;
	
	ITCSignal3() : m_vec() {}

	template <typename C>
	ConnectionType connect(C* inst, void (C::*call)(ArgType1, ArgType2, ArgType3))
	{
		return connect(boost::bind(call, inst, _1, _2, _3));
	}

	template <typename Caller, typename Callee = Caller>
	struct DoIt
	{
		typedef void (Callee::*CallType)(ArgType1, ArgType2, ArgType3);
		typedef void (Caller::*Type)(CallType, ArgType1, ArgType2, ArgType3);
	};
	
	template <typename Caller, typename Callee>
	ConnectionType connect(Caller* inst, 
						   typename DoIt<Caller, Callee>::CallType call)
	{
		return connect(boost::bind(typename DoIt<Caller,Callee>::Type(
										&Caller::template doIt<ArgType1, ArgType2, ArgType3>), 
								   inst, call, _1, _2, _3));
	}
	
	template <typename C>
	struct FuncDoIt
	{
		typedef void (*CallType)(ArgType1, ArgType2, ArgType3);
		typedef void (C::*Type)(CallType, ArgType1, ArgType2, ArgType3);
	};
	
	template <typename C>
	ConnectionType connect(C* inst, typename FuncDoIt<C>::CallType call)
	{
		return connect(boost::bind(typename FuncDoIt<C>::Type(
										&C::template doIt<ArgType1, ArgType2, ArgType3>), 
								   inst, call, _1, _2, _3));
	}
	
	ConnectionType connect(ConnectionType call)
	{
		m_vec.push_back(call);
		return m_vec[m_vec.size()-1];
	}
	
	void disconnect(ConnectionType call)
	{
		typename std::vector<SlotType>::iterator it = m_vec.begin();
		for (;it != m_vec.end(); ++it)
		{
			if (&(*it) == &call)
				break;
		}
		if (it != m_vec.end())
		{
			m_vec.erase(it);
		}
	}
	
	void operator()(ArgType1 arg1, ArgType2 arg2, ArgType3 arg3)
	{
		std::for_each(m_vec.begin(), m_vec.end(),
					  boost::bind(&SlotType::operator(), _1, arg1, arg2, arg3));
	}
	
private:
	std::vector<SlotType> m_vec;
};

};

#endif
